#ifndef SI_OBJECTS_H_
#define SI_OBJECTS_H_

/**
   @file

   This file defines the API for the following object services:
     - serialize database objects into a string;
     - materialize (deserialize) object from a string;
     - enumerating objects;
     - finding dependencies for objects;
     - executor for SQL statements;
     - wrappers for controlling the DDL Blocker;
*/

namespace obs {

/**
  Obj defines the basic set of operations for each database object.
*/

class Obj { 
public:

  bool serialize(THD *thd, String *serialialization);

  /**
    Return the name of the object.

    @return object name.
  */
  virtual const String *get_name() = 0;

  /**
    Return the database name of the object.

    @note this is a subject to remove.
  */
  virtual const String *get_db_name() = 0;

  bool execute(THD *thd);

public:
  virtual ~Obj()
  { }

private:

  /**
    Read the object state from a given buffer and restores object state to
    the point, where it can be executed.

    @param[in] serialialization_version The version of the serialization format.
    @param[in] serialialization         Buffer contained serialized object.

    @return error status.
      @retval FALSE on success.
      @retval TRUE on error.
  */
  virtual bool materialize(uint serialization_version,
                           const String *serialialization) = 0;

  /// Primitive implementing @c serialize() method.
  virtual bool do_serialize(THD *thd, String *serialialization) = 0;

  /// Primitive implementing @c execute() method.
  virtual bool do_execute(THD *thd) = 0;

  /**
    Drop the object.

    @param[in] thd              Server thread context.

    @return error status.
      @retval FALSE on success.
      @retval TRUE on error.
  */
  virtual bool drop(THD *thd) = 0;

private:
  friend Obj *materialize_database(const String *,
                                   uint,
                                   const String *);

  friend Obj *materialize_table(const String *,
                                const String *,
                                uint,
                                const String *);

  friend Obj *materialize_view(const String *,
                               const String *,
                               uint,
                               const String *);

  friend Obj *materialize_trigger(const String *,
                                  const String *,
                                  uint,
                                  const String *);

  friend Obj *materialize_stored_procedure(const String *,
                                           const String *,
                                           uint,
                                           const String *);

  friend Obj *materialize_stored_function(const String *,
                                          const String *,
                                          uint,
                                          const String *);

  friend Obj *materialize_event(const String *,
                                const String *,
                                uint,
                                const String *);

  friend Obj *materialize_tablespace(const String *,
                                     uint,
                                     const String *);

  friend Obj *materialize_db_grant(const String *,
                                   const String *,
                                   uint,
                                   const String *);

};


/**
  Create the object in the database.

  @param[in] thd              Server thread context.

  @return error status.
    @retval FALSE on success.
    @retval TRUE on error.

  @note The real work is done inside @c do_execute() primitive which should be
  defied in derived classes. This method prepares appropriate context and calls
  the primitive.
*/
inline
bool Obj::execute(THD *thd)
{
  ulong saved_sql_mode= thd->variables.sql_mode;
  thd->variables.sql_mode= 0;

  set_var_collation_client saved_charset_settings(
                             thd->variables.character_set_client,
                             thd->variables.character_set_results,
                             thd->variables.collation_connection);

  set_var_collation_client new_charset_settings(::system_charset_info,
                                                ::system_charset_info,
                                                ::system_charset_info);
  new_charset_settings.update(thd);

  bool ret= do_execute(thd);

  saved_charset_settings.update(thd);
  thd->variables.sql_mode= saved_sql_mode;

  return ret;
}

/**
  Serialize object state into a buffer. The buffer actually should be a
  binary buffer. String class is used here just because we don't have
  convenient primitive for binary buffers.

  Serialization format is opaque to the client, i.e. the client should
  not make any assumptions about the format or the content of the
  returned buffer.

  Serialization format can be changed in the future versions. However,
  the server must be able to materialize objects coded in any previous
  formats.

  @param[in] thd              Server thread context.
  @param[in] serialialization Buffer to serialize the object

  @return error status.
    @retval FALSE on success.
    @retval TRUE on error.

  @note The real work is done inside @c do_serialize() primitive which should be
  defied in derived classes. This method prepares appropriate context and calls
  the primitive.
*/
inline
bool Obj::serialize(THD *thd, String *serialization)
{
  ulong saved_sql_mode= thd->variables.sql_mode;
  thd->variables.sql_mode= 0;
  
  bool ret= do_serialize(thd, serialization);

  thd->variables.sql_mode= saved_sql_mode;

  return ret;
}

///////////////////////////////////////////////////////////////////////////

/**
  ObjIterator is a basic interface to enumerate the objects.
*/

class ObjIterator
{
public:

  ObjIterator()
  { }

  /**
    This operation returns a pointer to the next object in an enumeration.
    It returns NULL if there is no more objects.

    The client is responsible to destroy the returned object.

    @return a pointer to the object
      @retval NULL if there is no more objects in an enumeration.
  */
  virtual Obj *next() = 0;

public:
  virtual ~ObjIterator()
  { }

};

/**
  GrantObjIternator is an encapsulation of the three iterators for each level
  of grant supported: database-, table- and routine-, and column-level.
*/
class GrantObjIterator : public ObjIterator
{
public:
  GrantObjIterator(THD *thd, const String *db_name);

  ~GrantObjIterator() 
  {
    delete db_grants;
    delete tbl_grants;
    delete col_grants;
  }

  /**
    This operation returns a pointer to the next object in an enumeration.
    It returns NULL if there is no more objects.

    The client is responsible to destroy the returned object.

    @return a pointer to the object
      @retval NULL if there is no more objects in an enumeration.
  */
  Obj *next();

private:
  ObjIterator *db_grants;  ///< database-level grants
  ObjIterator *tbl_grants; ///< table- and routine-level grants
  ObjIterator *col_grants; ///< column-level grants
};

///////////////////////////////////////////////////////////////////////////

// The functions in this section are intended to construct an instance of
// Obj class for any particular database object. These functions do not
// interact with the server to validate requested names. So, it is possible
// to construct instances for non-existing objects.
//
// The client is responsible for destroying the returned object.

/**
  Construct an instance of Obj representing a database.

  No actual actions are performed in the server. An object can be created
  even for invalid database name or for non-existing database.

  The client is responsible to destroy the created object.

  @param[in] db_name Database name.

  @return a pointer to an instance of Obj representing given database.
*/

Obj *get_database(const String *db_name);

/**
  Construct an instance of Obj representing a table.

  No actual actions are performed in the server. An object can be created
  even for invalid database/table name or for non-existing table.

  The client is responsible to destroy the created object.

  @param[in] db_name    Database name.
  @param[in] table_name Table name.

  @return a pointer to an instance of Obj representing given table.
*/

Obj *get_table(const String *db_name, const String *table_name);

/**
  Construct an instance of Obj representing a view.

  No actual actions are performed in the server. An object can be created
  even for invalid database/view name or for non-existing view.

  The client is responsible to destroy the created object.

  @param[in] db_name   Database name.
  @param[in] view_name View name.

  @return a pointer to an instance of Obj representing given view.
*/

Obj *get_view(const String *db_name, const String *view_name);

/**
  Construct an instance of Obj representing a trigger.

  No actual actions are performed in the server. An object can be created
  even for invalid database/trigger name or for non-existing trigger.

  The client is responsible to destroy the created object.

  @param[in] db_name      Database name.
  @param[in] trigger_name Trigger name.

  @return a pointer to an instance of Obj representing given trigger.
*/

Obj *get_trigger(const String *db_name, const String *trigger_name);

/**
  Construct an instance of Obj representing a stored procedure.

  No actual actions are performed in the server. An object can be created
  even for invalid database/procedure name or for non-existing stored
  procedure.

  The client is responsible to destroy the created object.

  @param[in] db_name Database name.
  @param[in] sp_name Stored procedure name.

  @return a pointer to an instance of Obj representing given stored
  procedure.
*/

Obj *get_stored_procedure(const String *db_name, const String *sp_name);

/**
  Construct an instance of Obj representing a stored function.

  No actual actions are performed in the server. An object can be created
  even for invalid database/function name or for non-existing stored
  function.

  The client is responsible to destroy the created object.

  @param[in] db_name Database name.
  @param[in] sf_name Stored function name.

  @return a pointer to an instance of Obj representing given stored
  function.
*/

Obj *get_stored_function(const String *db_name, const String *sf_name);

/**
  Construct an instance of Obj representing an event.

  No actual actions are performed in the server. An object can be created
  even for invalid database/event name or for non-existing event.

  The client is responsible to destroy the created object.

  @param[in] db_name    Database name.
  @param[in] event_name Event name.

  @return a pointer to an instance of Obj representing given event.
*/

Obj *get_event(const String *db_name, const String *event_name);

///////////////////////////////////////////////////////////////////////////

// The functions in this section provides a way to iterator over all
// objects in the server or in the particular database.
//
// The client is responsible for destroying the returned iterator.

/**
  Create an iterator over all databases in the server.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
*/

ObjIterator *get_databases(THD *thd);

/**
  Create an iterator over all tables in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_tables(THD *thd, const String *db_name);

/**
  Create an iterator over all views in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_views(THD *thd, const String *db_name);

/**
  Create an iterator over all triggers in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_triggers(THD *thd, const String *db_name);

/**
  Create an iterator over all stored procedures in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_stored_procedures(THD *thd, const String *db_name);

/**
  Create an iterator over all stored functions in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_stored_functions(THD *thd, const String *db_name);

/**
  Create an iterator over all events in the particular database.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator *get_db_events(THD *thd, const String *db_name);

/*
  Creates a high-level iterator that iterates over database-, table-,
  routine-, and column-level privileges which shall permit a single
  iterator from the si_objects to retrieve all of the privileges for 
  a given database.
*/
ObjIterator *get_all_db_grants(THD *thd, const String *db_name);

///////////////////////////////////////////////////////////////////////////

// The functions are intended to enumerate dependent objects.
//
// The client is responsible for destroying the returned iterator.

/**
  Create an iterator overl all base tables in the particular view.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator* get_view_base_tables(THD *thd,
                                  const String *db_name,
                                  const String *view_name);

/**
  Create an iterator overl all base tables in the particular view.

  The client is responsible to destroy the returned iterator.

  @return a pointer to an iterator object.
    @retval NULL in case of error.
*/

ObjIterator* get_view_base_views(THD *thd,
                                 const String *db_name,
                                 const String *view_name);

///////////////////////////////////////////////////////////////////////////

// The functions in this section provides a way to materialize objects from
// the serialized form.
//
// The client is responsible for destroying the returned iterator.

Obj *materialize_database(const String *db_name,
                          uint serialization_version,
                          const String *serialialization);

Obj *materialize_table(const String *db_name,
                       const String *table_name,
                       uint serialization_version,
                       const String *serialialization);

Obj *materialize_view(const String *db_name,
                      const String *view_name,
                      uint serialization_version,
                      const String *serialialization);

Obj *materialize_trigger(const String *db_name,
                         const String *trigger_name,
                         uint serialization_version,
                         const String *serialialization);

Obj *materialize_stored_procedure(const String *db_name,
                                  const String *stored_proc_name,
                                  uint serialization_version,
                                  const String *serialialization);

Obj *materialize_stored_function(const String *db_name,
                                 const String *stored_func_name,
                                 uint serialization_version,
                                 const String *serialialization);

Obj *materialize_event(const String *db_name,
                       const String *event_name,
                       uint serialization_version,
                       const String *serialialization);

Obj *materialize_tablespace(const String *ts_name,
                            uint serialization_version,
                            const String *serialialization);

Obj *materialize_db_grant(const String *grantee,
                          const String *db_name,
                          uint serialization_version,
                          const String *serialization);

///////////////////////////////////////////////////////////////////////////

bool is_internal_db_name(const String *db_name);

///////////////////////////////////////////////////////////////////////////

/**
  Check if the given directory actually exists.

  @return Error status.
    @retval FALSE on success (the database exists and accessible).
    @retval TRUE on error (the database either not exists, or not accessible).
*/
bool check_db_existence(const String *db_name);

/*
  Returns TRUE if user is defined on the system.
*/
bool user_exists(THD *thd, const String *grantee);

/*
  This method returns a @c TablespaceObj object if the table has a tablespace.
*/
Obj *get_tablespace_for_table(THD *thd, 
                              const String *db_name, 
                              const String *tbl_name);

/*
  This method determines if a materialized tablespace exists on the
  system. This compares the name and all saved attributes of the 
  tablespace. A FALSE return would mean either the tablespace does
  not exist or the tablespace attributes are different.
*/
bool tablespace_exists(THD *thd,
                       Obj *ts);

/*
  This method determines if the tablespace referenced by name exists on the
  system. Returns a TablespaceObj if it exists or NULL if it doesn't.
*/
Obj *is_tablespace(THD *thd,
                   const String *ts_name);

/*
  This method returns a description of the tablespace useful for communicating
  with the user.
*/
const String *describe_tablespace(Obj *ts);

///////////////////////////////////////////////////////////////////////////

//
// DDL blocker methods.
//

/**
  Turn on the ddl blocker.

  This method is used to start the ddl blocker blocking DDL commands.

  @param[in] thd  Thread context.

  @return error status.
    @retval FALSE on success.
    @retval TRUE on error.
*/
bool ddl_blocker_enable(THD *thd);

/**
  Turn off the ddl blocker.

  This method is used to stop the ddl blocker from blocking DDL commands.
*/
void ddl_blocker_disable();

/**
  Turn on the ddl blocker exception

  This method is used to allow the exception allowing a restore operation to
  perform DDL operations while the ddl blocker blocking DDL commands.

  @param[in] thd  Thread context.
*/
void ddl_blocker_exception_on(THD *thd);

/**
  Turn off the ddl blocker exception.

  This method is used to suspend the exception allowing a restore operation to
  perform DDL operations while the ddl blocker blocking DDL commands.

  @param[in] thd  Thread context.
*/
void ddl_blocker_exception_off(THD *thd);

/*
  Creates a WHERE clause for information schema table lookups of the
  for FROM INFORMATION_SCHEMA.X WHERE <db_col> IN ('a','b','c').
*/
COND *create_db_select_condition(THD *thd, 
                                 TABLE *t,
                                 List<LEX_STRING> *db_list);

/*
  The following class is used to manage name locks on a list of tables.

  This class uses a list of type List<Obj> to establish the table list 
  that will be used to manage locks on the tables.
*/
class Name_locker
{
public:
  Name_locker(THD *thd) { m_thd= thd; }
  ~Name_locker() 
  { 
    free_table_list(m_table_list);
    m_table_list= NULL;
  }

  /*
    Gets name locks on table list.
  */
  int get_name_locks(List<Obj> *tables, thr_lock_type lock);

  /*
    Releases name locks on table list.
  */
  int release_name_locks();

private:
  TABLE_LIST *m_table_list; ///< The list of tables to obtain locks on.
  THD *m_thd;               ///< Thread context.

  /*
    Builds a table list from the list of objects passed to constructor.
  */
  TABLE_LIST *build_table_list(List<Obj> *tables, thr_lock_type lock);
  void free_table_list(TABLE_LIST*);
};

} // obs namespace

#endif // SI_OBJECTS_H_
